import logging
from typing import Any, Dict, Iterator, List, Tuple

import numpy as np
from numpy.typing import NDArray
from scipy.special import sph_harm

from ...core.hashing import list_to_hash
from .base_basis_set import BasisSet
from ... import ProbedCoordinates


logger = logging.getLogger(__name__)


class SphericalHarmonics(BasisSet):
    """ Basis set class for spherical harmonics, the canonical representation
    of polynomials on the unit sphere and a simple way of representing
    band-limited spherical functions which allows for easy computations of statistics
    and is suitable for analyzing certain symmetries.

    Parameters
    ----------
    ell_max : int
        The bandlimit of the spherical functions that you want to be able to represent.
        A good rule of thumb is that :attr:`ell_max` should not exceed the
        number of detector segments minus 1.
    probed_coordinates : ProbedCoordinates
        Optional. A container with the coordinates on the sphere probed at each detector segment by the
        experimental method. Its construction from the system geometry is method-dependent.
        By default, an empty instance of :class:`mumott.ProbedCoordinates` is created.
    """
    def __init__(self,
                 probed_coordinates: ProbedCoordinates = None,
                 ell_max: int = 0):
        if probed_coordinates is None:
            probed_coordinates = ProbedCoordinates()
        self._probed_coordinates = probed_coordinates
        self._probed_coordinates_hash = hash(probed_coordinates)
        self._ell_max = ell_max
        self._ell_indices = np.zeros(1)
        self._emm_indices = np.zeros(1)
        self._projection_matrix = np.zeros((self._probed_coordinates.vector.shape[0],
                                           self._probed_coordinates.vector.shape[1],
                                           len(self._ell_indices)), dtype=np.float64)
        # Compute initial values for indices and matrix.
        self._get_coefficient_indices()
        self._get_projection_matrix()

    def _get_coefficient_indices(self) -> None:
        """
        Computes the attributes :attr:`~.SphericalHarmonics.ell_indices` and
        :attr:`~.SphericalHarmonics.emm_indices`. Called when :attr:`~.SphericalHarmonics.ell_max`
        changes.
        """
        mm = np.zeros((self._ell_max + 1) * (self._ell_max // 2 + 1), dtype=int)
        ll = np.zeros((self._ell_max + 1) * (self._ell_max // 2 + 1), dtype=int)
        count = 0
        for h in range(0, self._ell_max + 1, 2):
            for i in range(-h, h + 1):
                ll[count] = h
                mm[count] = i
                count += 1
        self._ell_indices = ll
        self._emm_indices = mm

    def _get_projection_matrix(self) -> None:
        """ Computes the matrix necessary for forward and gradient calculations.
        Called when the coordinate system has been updated or ``ell_max`` has changed."""
        _, probed_polar_angles, probed_azim_angles = self._probed_coordinates.to_spherical
        # retrieve complex spherical harmonics with emm >= 0, shape (N, M, len(ell_indices), K)
        complex_factors = sph_harm(abs(self._emm_indices)[np.newaxis, np.newaxis, ..., np.newaxis],
                                   self._ell_indices[np.newaxis, np.newaxis, ..., np.newaxis],
                                   probed_azim_angles[..., np.newaxis, :],
                                   probed_polar_angles[..., np.newaxis, :])
        # reduce over K
        complex_factors = complex_factors.mean(axis=-1)
        # cancel Condon-Shortley phase factor in scipy.special.sph_harm
        condon_shortley_factor = (-1.) ** self._emm_indices
        # 4pi normalization factor and complex-to-real normalization factor for m != 0
        norm_factor = np.sqrt(4 * np.pi) * \
            np.sqrt(1 + (self._emm_indices != 0).astype(int)) * condon_shortley_factor
        self._projection_matrix = norm_factor[np.newaxis, np.newaxis, ...] * (
            (self._emm_indices >= 0)[np.newaxis, np.newaxis, ...] * complex_factors.real +
            (self._emm_indices < 0)[np.newaxis, np.newaxis, ...] * complex_factors.imag)

    def forward(self,
                coefficients: NDArray,
                indices: NDArray = None) -> NDArray:
        """ Carries out a forward computation of projections from spherical harmonic space
        into detector space, for one or several tomographic projections.

        Parameters
        ----------
        coefficients
            An array of coefficients, of arbitrary shape so long as the last
            axis has the same size as :attr:`~.SphericalHarmonics.ell_indices`, and if
            :attr:`indices` is `None` or greater than one, the first axis should have the
            same length as :attr:`indices`
        indices
            Optional. Indices of the tomographic projections for which the forward
            computation is to be performed. If ``None``, the forward computation will
            be performed for all projections.

        Returns
        -------
            An array of values on the detector corresponding to the :attr:`coefficients` given.
            If :attr:`indices` contains exactly one index, the shape is ``(coefficients.shape[:-1], J)``
            where ``J`` is the number of detector segments. If :attr:`indices` is ``None`` or contains
            several indices, the shape is ``(N, coefficients.shape[1:-1], J)`` where ``N``
            is the number of tomographic projections for which the computation is performed.

        Notes
        -----
        The assumption is made in this implementation that computations over several
        indices act on sets of images from different projections. For special usage
        where multiple projections of entire fields is desired, it may be better
        to use :attr:`projection_matrix` directly. This also applies to
        :meth:`gradient`.
        """
        assert coefficients.shape[-1] == self._ell_indices.size
        self._update()
        if indices is None:
            return np.einsum('ijk, i...k -> i...j',
                             self._projection_matrix,
                             coefficients,
                             optimize='greedy')
        if indices.size == 1:
            return np.einsum('ijk, ...k -> ...j',
                             self._projection_matrix[indices],
                             coefficients,
                             optimize='greedy')
        np.einsum('ijk, i...k -> i...j',
                  self._projection_matrix[indices],
                  coefficients,
                  optimize='greedy')

    def gradient(self,
                 coefficients: NDArray,
                 indices: NDArray = None) -> NDArray:
        """ Carries out a gradient computation of projections from spherical harmonic space
        into detector space, for one or several tomographic projections.

        Parameters
        ----------
        coefficients
            An array of coefficients (or residuals) of arbitrary shape so long as the last
            axis has the same size as the number of detector segments.
        indices
            Optional. Indices of the tomographic projections for which the gradient
            computation is to be performed. If ``None``, the gradient computation will
            be performed for all projections.

        Returns
        -------
            An array of gradient values based on the `coefficients` given.
            If :attr:`indices` contains exactly one index, the shape is ``(coefficients.shape[:-1], J)``
            where ``J`` is the number of detector segments. If indices is ``None`` or contains
            several indices, the shape is ``(N, coefficients.shape[1:-1], J)`` where ``N``
            is the number of tomographic projections for which the computation is performed.

        Notes
        -----
        When solving an inverse problem, one should not to attempt to optimize the
        coefficients directly using the ``gradient`` one obtains by applying this method to the data.
        Instead, one must either take the gradient of the residual between the
        :meth:`~.SphericalHarmonics.forward` computation of the coefficients and the data.
        Alternatively one can apply both the forward and the gradient computation to the
        coefficients to be optimized, and the gradient computation to the data, and treat
        the residual of the two as the gradient of the optimization coefficients. The approaches
        are algebraically equivalent, but one may be more efficient than the other in some
        circumstances.
        """
        self._update()
        if indices is None:
            return np.einsum('ikj, i...k -> i...j',
                             self._projection_matrix,
                             coefficients,
                             optimize='greedy')
        if indices.size == 1:
            # index logic is a little weird here because indices is an array
            return np.einsum('ikj, ...k -> ...j',
                             self._projection_matrix[indices],
                             coefficients,
                             optimize='greedy')
        return np.einsum('ikj, i...k -> i...j',
                         self._projection_matrix[indices],
                         coefficients,
                         optimize='greedy')

    def get_inner_product(self,
                          u: NDArray,
                          v: NDArray,
                          resolve_spectrum: bool = False,
                          spectral_moments: List[int] = None) -> NDArray:
        r""" Retrieves the inner product of two coefficient arrays.

        Notes
        -----
            The canonical inner product in a spherical harmonic representation
            is :math:`\sum_\ell N(\ell) \sum_m u_m^\ell v_m^\ell`, where :math:`N(\ell)` is
            a normalization constant (which is unity for the :math:`4\pi` normalization).
            This inner product is a rotational invariant. The rotational invariance also holds
            for any partial sums over :math:`\ell`.
            One can define a function of :math:`\ell` that returns such
            products, namely :math:`S(\ell, u, v) = N(\ell)\sum_m u_m^\ell v_m^\ell`,
            called the spectral power function.
            The sum :math:`\sum_{\ell = 1}S(\ell)` is equal to the covariance of the
            band-limited spherical functions represented by :math:`u` and :math:`v`, and each
            :math:`S(\ell, u, v)` is the contribution to the covariance of the band :math:`\ell`.
            See also
            `the SHTOOLS documentation <https://shtools.github.io/SHTOOLS/real-spherical-harmonics.html>`_
            for an excellent overview of this.

        Parameters
        ----------
        u
            The first coefficient array, of arbitrary shape and dimension,
            except the last dimension must be the same as the length of
            :attr:`~.SphericalHarmonics.ell_indices`.
        v
            The second coefficient array, of the same shape as ``u``.
        resolve_spectrum
            Optional. Whether to resolve the product according to each frequency band, given by the
            coefficients of each ``ell`` in :attr:`~SphericalHarmonics.ell_indices`.
            Defaults to ``False``, which means that the sum of every component of the spectrum is returned.
            If ``True``, components are returned in order of ascending ``ell``.
            The ``ell`` included in the spectrum depends on :attr:`spectral_moments`.
        spectral_moments
            Optional. List of particular values of ``ell`` to calculate the inner product for.
            Defaults to ``None``, which is identical to including all values of ``ell``
            in the calculation.
            If :attr:`spectral_moments` contains all nonzero values of ``ell`` and
            :attr:`resolve_spectrum` is ``False``, the covariance of
            :attr:`v` and :attr:`u` will be calculated (the sum of the inner product over all non-zero ``ell``
            If ``resolve_spectrum`` is ``True``, the covariance per `ell` in ``spectral_moments``, will
            be calculated, i.e., the inner products will not be summed over.

        Returns
        -------
            An array of the inner products of the spherical functions represented by ``u`` and ``v``.
            Has the shape ``(u.shape[:-1])`` if :attr:`resolve_spectrum` is ``False``,
            ``(u.shape[:-1] + (ell_max // 2 + 1,))`` if :attr:`resolve_spectrum` is ``True`` and
            ``spectral_moments`` is ``None``, and finally the shape
            ``(u.shape[:-1] + (np.unique(spectral_moments).size,))`` if :attr:`resolve_spectrum` is ``True``
            and :attr:`spectral_moments` is a list of integers found
            in :attr:`~.SphericalHarmonics.ell_indices`
         """
        assert u.shape == v.shape
        assert u.shape[-1] == self._ell_indices.size
        if not resolve_spectrum:
            if spectral_moments is None:
                return np.einsum('...i, ...i -> ...', u, v)
            # pick out only the subset where ell matches the provided spectral moments
            where = np.any([np.equal(self._ell_indices, ell) for ell in spectral_moments], axis=0)
            return np.einsum('...i, ...i -> ...', u[..., where], v[..., where])
        if spectral_moments is None:
            which_ell = np.unique(self._ell_indices)
        else:
            which_ell = np.unique(spectral_moments)
            which_ell = [ell for ell in which_ell if np.any(np.equals(self._ell_indices, ell))]
        power_spectrum = np.zeros((*u.shape[:-1], which_ell.size))
        # power spectrum for any one ell is given by inner product over each ell
        for i, ell in enumerate(which_ell):
            power_spectrum[..., i] = np.einsum('...i, ...i -> ...',
                                               u[..., self._ell_indices == ell],
                                               v[..., self._ell_indices == ell],
                                               optimize='greedy')
        return power_spectrum

    def get_covariances(self,
                        u: NDArray,
                        v: NDArray,
                        resolve_spectrum: bool = False) -> NDArray:
        """ Returns the covariances of the spherical functions represented by
        two coefficient arrays.

        Parameters
        ----------
        u
            The first coefficient array, of arbitrary shape
            except its last dimension must be the same length as
            the length of :attr:`~SphericalHarmonics.ell_indices`.
        v
            The second coefficient array, of the same shape as :attr:`u`.
        resolve_spectrum
            Optional. Whether to resolve the product according to each frequency band, given by the
            coefficients of each ``ell`` in :attr:`~.SphericalHarmonics.ell_indices`.
            Default value is ``False``.

        Returns
        -------
            An array of the covariances of the spherical functions represented by ``u`` and ``v``.
            Has the shape ``(u.shape[:-1])`` if `resolve_spectrum` is ``False``, and
            ``(u.shape[:-1] + (ell_max // 2 + 1,))`` if `resolve_spectrum` is ``True``, where
            ``ell_max`` is :attr:`.SphericalHarmonics.ell_max`.

        Notes
        -----
        Calling this function is equivalent to calling :func:`~.SphericalHarmonics.get_inner_product`
        with ``spectral_moments=np.unique(ell_indices[ell_indices > 0])`` where ``ell_indices`` is
        :attr:`.SphericalHarmonics.ell_indices`. See the note to
        :func:`~.SphericalHarmonics.get_inner_product` for mathematical details.
        """
        spectral_moments = np.unique(self._ell_indices[self._ell_indices > 0])
        return self.get_inner_products(u, v, resolve_spectrum, spectral_moments)

    def get_output(self,
                   coefficients: NDArray) -> Dict[str, Any]:
        r""" Returns a dictionary of output data for a given array of spherical harmonic coefficients.

        Parameters
        ----------
        coefficients
            An array of coefficients of arbitrary shape and dimensions, except
            its last dimension must be the same length as :attr:`~.SphericalHarmonics.ell_indices`.
            Computations only operate over the last axis of :attr:`coefficents`, so derived
            properties in the output will have the shape ``(*coefficients.shape[:-1], ...)``.

        Returns
        -------
            A dictionary containing two sub-dictionaries, ``basis_set`` and ``spherical_functions``.
            ``basis_set`` contains information particular to :class:`SphericalHarmonics`, whereas
            ``spherical_functions`` contains information about the spherical functions
            represented by the :attr:`coefficients` which are not specific to the chosen representation.

        Notes
        -----
        In detail, the two sub-dictionaries ``basis_set`` and ``spherical_functions`` have the following
        members:

        basis_set
            name
                The name of the basis set, i.e., ``'SphericalHarmonicParameters'``
            coefficients
                A copy of :attr:`coefficients`.
            ell_max
                A copy of :attr:`~.SphericalHarmonics.ell_max`.
            ell_indices
                A copy of :attr:`~.SphericalHarmonics.ell_indices`.
            emm_indices
                A copy of :attr:`~.SphericalHarmonics.emm_indices`.
            projection_matrix
                A copy of :attr:`~.SphericalHarmonics.projection_matrix`.
        spherical_functions
            means
                The spherical means of each function represented by :attr:`coefficients`.
            variances
                The spherical variances of each function represented by :attr:`coefficients`.
                If :attr:`~.ell_max` is ``0``, all variances will equal zero.
            r2_tensors
                The traceless symmetric rank-2 tensor component of each function represented by
                :attr:`coefficients`, in 6-element form, in the order ``[xx, yy, zz, yz, xz, xy]``,
                i.e., by the Voigt convention.
                The matrix form can be recovered as r2_tensors[..., tensor_to_matrix_indices],
                yielding matrix elements ``[[xx, xy, xz], [xy, yy, yz], [xz, yz, zz]]``.
                If :attr:`~.ell_max` is ``0``, all tensors have elements
                [1, 0, -1, 0, 0, 0].
            tensor_to_matrix_indices
                A list of indices to help recover the matrix from the 2-element form of the
                rank-2 tensors, equalling precisely ``[[0, 5, 4], [5, 1, 3], [4, 3, 2]]``
            eigenvalues
                The eigenvalues of the rank-2 tensors, sorted in ascending order in the last index.
                If :attr:`~.ell_max` is ``0``, the eigenvalues will always be (1, 0, -1)
            eigenvectors
                The eigenvectors of the rank-2 tensors, sorted with their corresponding eigenvectors
                in the last index. Thus, ``eigenvectors[..., 0]`` gives the eigenvector corresponding
                to the smallest eigenvalue, and ``eigenvectors[..., 2]`` gives the eigenvector
                corresponding to the largest eigenvalue. Generally, one of these two eigenvectors
                is used to define the orientation of a function, depending on whether it is
                characterized by a minimum (``0``) or a maximum (``2``). The middle eigenvector (``1``)
                is typically only used for visualizations.
                If :attr:`~.ell_max` is ``0``, the eigenvectors will be the Cartesian basis
                vectors.
            main_orientations
                The estimated main orientations from the largest absolute eigenvalues.
                If :attr:`~.ell_max` is ``0``, the main orientation will be the x-axis.
            main_orientation_symmetries
                The strength or definiteness of the main orientation, calculated from
                the quotient of the absolute middle and signed largest eigenvalues of the
                rank-2 tensor.
                If ``0``, the orientation is totally ambiguous. The orientation is completely
                transversal if the value is ``-1`` (orientation represents a minimum),
                and completely longitudinal if the value is ``1`` (orientation represents a maximum).
                If :attr:`~.ell_max` is ``0``, the main orientations are all totally ambiguous.
            normalized_standard_deviations
                A relative measure of the overall anisotropy of the spherical functions. Equals
                :math:`\sqrt{\sigma^2 / \mu}`, where :math:`\sigma^2` is the variance
                and :math:`\mu` is the mean. The places where :math:`\mu=0` have been
                set to ``0``. If :attr:`~.ell_max` is ``0``, the normalized standard
                deviations will always be zero.
            power_spectra
                The spectral powers of each ``ell`` in :attr:`~.SphericalHarmonics.ell_indices`,
                for each spherical function, sorted in ascending ``ell``.
                If :attr:`~.ell_max` is ``0``, each function will have only one element, equal
                to the mean squared.
            power_spectra_ell
                An array containing the corresponding ``ell`` to each of the last indices
                in :attr:`power_spectra`. Equal to ``np.unique(ell_indices)``.
        """
        assert coefficients.shape[-1] == self._ell_indices.size
        # Update to ensure non-dirty output state.
        self._update()
        output_dictionary = {}

        # basis set-specific information
        basis_set = {}
        output_dictionary['basis_set'] = basis_set
        basis_set['name'] = type(self).__name__
        basis_set['coefficients'] = coefficients.copy()
        basis_set['ell_max'] = self._ell_max
        basis_set['ell_indices'] = self._ell_indices.copy()
        basis_set['emm_indices'] = self._emm_indices.copy()
        basis_set['projection_matrix'] = self._projection_matrix.copy()
        basis_set['hash'] = hex(hash(self))

        # general properties of spherical function
        spherical_functions = {}
        output_dictionary['spherical_functions'] = spherical_functions
        spherical_functions['means'] = coefficients[..., 0]
        if coefficients.shape[-1] > 1:
            spherical_functions['variances'] = (coefficients[..., 1:] ** 2).sum(-1)
        else:
            spherical_functions['variances'] = np.zeros_like(coefficients[..., 0])
        r2_tensor, eigvect, eigval = self._get_rank2_tensor_analysis(coefficients)
        spherical_functions['eigenvectors'] = eigvect
        spherical_functions['r2_tensors'] = np.concatenate(
             (np.diagonal(r2_tensor, offset=0, axis1=-2, axis2=-1),
              r2_tensor[..., 1, [2]], r2_tensor[..., 0, [2]], r2_tensor[..., 0, [1]]), axis=-1)
        spherical_functions['tensor_to_matrix_indices'] = [[0, 5, 4], [5, 1, 3], [4, 3, 2]]
        spherical_functions['eigenvalues'] = eigval

        # estimate main orientation from absolute eigenvalues and select from eigenvectors
        spherical_functions['main_orientations'] = \
            np.take_along_axis(eigvect,
                               np.argmax(abs(eigval), axis=-1).reshape(eigval.shape[:-1] + (1, 1)), axis=-1)
        spherical_functions['main_orientations'] = spherical_functions['main_orientations'][..., 0]
        # estimate main orientation strength as quotient between middle and largest absolute eigenvalue.
        eigargs = np.argmax(abs(eigval), axis=-1).reshape(eigval.shape[:-1] + (1,))
        spherical_functions['main_orientation_symmetries'] = 2 * abs(eigval[..., 1]) / \
            np.take_along_axis(eigval, eigargs, axis=-1)

        # estimate overall relative anisotropy as quotient betewen standard deviation and mean
        valid_indices = spherical_functions['means'] > 0  # mask points where mean is zero or negative
        normalized_std = np.zeros_like(spherical_functions['means'])
        normalized_std[valid_indices] = np.sqrt(spherical_functions['variances'][valid_indices]) / \
            spherical_functions['means'][valid_indices]
        spherical_functions['normalized_standard_deviations'] = normalized_std
        spherical_functions['power_spectra'] = self.get_inner_product(coefficients,
                                                                      coefficients,
                                                                      resolve_spectrum=True)
        spherical_functions['power_spectra_ell'] = np.unique(self._ell_indices)
        return output_dictionary

    def _get_rank2_tensor_analysis(self, coefficients: NDArray) -> Tuple[NDArray, NDArray, NDArray]:
        """ Performs an analysis of the rank-2 tensor components of the functions represented by the
        given coefficients.

        Parameters
        ----------
        coefficients
            An array of coefficients of arbitrary shape, as long as the last axis has the same shape
            as :attr:`~.SphericalHarmonics.ell_indices`.

        Returns
        -------
            A tuple with three memberrs.
            First, the traceless rank-2 tensor components of all the
            functions represented by the :attr:`coefficients`, in 3-by-3 matrix form stored
            in the last two indices: ``[[xx, xy, xz], [xy, yy, yz], [xz, yz, zz]]``
            Second, eigenvalues of the rank-2 tensors, sorted in ascending order.
            Third, eigenvectors associated to each eigenvalue, sorted in the same order in the last index.
            Thus, ``eigenvectors[..., 0]`` returns the eigenvectors corresponding to the smallest eigenvalue.

        Notes
        -----
        This method handles the edge case where :attr:`~.SphericalHarmonics.ell_max` is ``0`` gracefully.
        In this case, all elements of :attr:`r2_tensors` will be ``[[1, 0, 0], [0, 0, 0], [0, 0, -1]]``,
        all elements of :attr:`eigenvalues` will be ``[-1, 0, 1]``, and all elements of
        :attr:`eigenvectors` will ``[[1, 0, 0], [0, 1, 0], [0, 0, 1]]``.

        """
        if self._ell_max < 2:
            logger.info('Note: ell_max < 2, so rank-2 tensors will all be diagonal matrices with diagonal'
                        ' [1, 0, -1], eigenvalues will be [1, 0, -1],'
                        ' and eigenvectors will be Cartesian basis vectors.')
            eigenvalues = np.zeros(coefficients.shape[:-1] + (3,), dtype=coefficients.dtype)
            eigenvectors = np.zeros(coefficients.shape[:-1] + (3, 3), dtype=coefficients.dtype)
            eigenvalues[...] = [-1, 0, 1]
            eigenvectors[..., 0, 0] = 1.
            eigenvectors[..., 1, 1] = 1.
            eigenvectors[..., 2, 2] = 1.
            r2_tensor = np.zeros(coefficients.shape[:-1] + (3, 3), dtype=coefficients.dtype)
            r2_tensor[..., 0, 0] = 1
            r2_tensor[..., 1, 1] = 0
            r2_tensor[..., 2, 2] = -1
            return r2_tensor, eigenvectors, eigenvalues
        A = coefficients[..., self._ell_indices == 2]

        # Normalizing coefficients
        c1 = np.sqrt(15)
        c2 = np.sqrt(5)
        c3 = np.sqrt(15)
        r2_tensor = np.zeros(coefficients.shape[:-1] + (3, 3), dtype=coefficients.dtype)
        r2_tensor[..., 0, 0] = c3 * A[..., 4] - c2 * A[..., 2]
        r2_tensor[..., 1, 1] = -c3 * A[..., 4] - c2 * A[..., 2]
        r2_tensor[..., 2, 2] = c2 * 2 * A[..., 2]
        r2_tensor[..., 0, 1] = c1 * A[..., 0]
        r2_tensor[..., 1, 0] = c1 * A[..., 0]
        r2_tensor[..., 2, 1] = c1 * A[..., 1]
        r2_tensor[..., 1, 2] = c1 * A[..., 1]
        r2_tensor[..., 2, 0] = c1 * A[..., 3]
        r2_tensor[..., 0, 2] = c1 * A[..., 3]
        w, v = np.linalg.eigh(r2_tensor.reshape(-1, 3, 3))

        # Some complicated sorting logic to sort eigenvectors per ascending eigenvalues.
        sorting = np.argsort(w, axis=1).reshape(-1, 3, 1)
        v = v.transpose(0, 2, 1)
        v = np.take_along_axis(v, sorting, axis=1)
        v = v.transpose(0, 2, 1)
        v = v / np.sqrt(np.sum(v ** 2, axis=1).reshape(-1, 1, 3))
        eigenvalues = w.reshape(coefficients.shape[:-1] + (3,))
        eigenvectors = v.reshape(coefficients.shape[:-1] + (3, 3,))
        return r2_tensor.reshape(*coefficients.shape[:-1], 3, 3), eigenvectors, eigenvalues

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        """ Allows class to be iterated over and in particular be cast as a dictionary.
        """
        yield 'name', type(self).__name__
        yield 'ell_max', self._ell_max
        yield 'ell_indices', self._ell_indices
        yield 'emm_indices', self._emm_indices
        yield 'projection_matrix', self._projection_matrix
        yield 'hash', hex(hash(self))[2:]

    def __len__(self) -> int:
        return len(self._ell_indices)

    def __hash__(self) -> int:
        """Returns a hash reflecting the internal state of the instance.

        Returns
        -------
            A hash of the internal state of the instance,
            cast as an ``int``.
        """
        to_hash = [self._ell_max,
                   self._ell_indices,
                   self._emm_indices,
                   self._projection_matrix,
                   self._probed_coordinates_hash]
        return int(list_to_hash(to_hash), 16)

    def _update(self) -> None:
        # We only run updates if the hashes do not match.
        if self.is_dirty:
            self._get_projection_matrix()
            self._probed_coordinates_hash = hash(self._probed_coordinates)

    @property
    def is_dirty(self) -> bool:
        return hash(self._probed_coordinates) != self._probed_coordinates_hash

    @property
    def projection_matrix(self) -> NDArray:
        """ The matrix used to project spherical functions from the unit sphere onto the detector.
        If ``v`` is a vector of spherical harmonic coefficients, and ``M`` is the ``projection_matrix``,
        then ``M @ v`` gives the corresponding values on the detector segments associated with
        each projection. ``M[i] @ v`` gives the values on the detector segments associated with
        projection ``i``.

        If ``r`` is a residual between a projection from spherical to detector space and data from
        projection ``i``, then ``M[i].T @ r`` gives the associated gradient in spherical harmonic
        space.
        """
        self._update()
        return self._projection_matrix

    @property
    def ell_max(self) -> int:
        r""" The maximum ``ell`` used to represent spherical functions.

        Notes
        -----
        The word ``ell`` is used to represent the cursive small L, also written
        :math:`\ell`, often used as an index for the degree of the Legendre polynomial
        in the definition of the spherical harmonics.
        """
        return self._ell_max

    @ell_max.setter
    def ell_max(self, val: int) -> NDArray[int]:
        if (val % 2 != 0) or (val < 0) or (val != round(val)):
            raise ValueError('ell_max must be an even, non-negative integer,'
                             f' but a value of {self._ell_max} was given!')
        self._ell_max = val
        self._get_coefficient_indices()
        self._get_projection_matrix()

    @property
    def ell_indices(self) -> NDArray[int]:
        r""" The ``ell`` associated with each coefficient and its corresponding
        spherical harmonic. Updated when :attr:`~.SphericalHarmonics.ell_max` changes.

        Notes
        -----
        The word ``ell`` is used to represent the cursive small L, also written
        :math:`\ell`, often used as an index for the degree of the Legendre polynomial
        in the definition of the spherical harmonics.
        """
        return self._ell_indices

    @property
    def emm_indices(self) -> NDArray[int]:
        r""" The ``emm`` associated with each coefficient and its corresponding
        spherical harmonic. Updated when :attr:`~.SphericalHarmonics.ell_max` changes.

        Notes
        -----
        For consistency with :attr:`~.SphericalHarmonics.ell_indices`, and to avoid
        visual confusion with other letters, ``emm`` is
        used to represent the index commonly written :math:`m` in mathematical notation,
        the frequency of the sine-cosine parts of the spherical harmonics,
        often called the spherical harmonic order.
        """
        return self._emm_indices

    @property
    def probed_coordinates(self) -> ProbedCoordinates:
        """ The instance of :class:`mumott.ProbedCoordinates` attached to this
        object. If modified via the :attr:`mumott.ProbedCoordinates.vector`
        attribute, the :class:`mumott.methods.SphericalHarmonics` instance is
        automatically updated when needed."""
        return self._probed_coordinates

    def __str__(self) -> str:
        wdt = 74
        s = []
        s += ['-' * wdt]
        s += ['SphericalHarmonics'.center(wdt)]
        s += ['-' * wdt]
        with np.printoptions(threshold=4, edgeitems=2, precision=5, linewidth=60):
            s += ['{:18} : {}'.format('Maximum "ell"', self.ell_max)]
            s += ['{:18} : {}'.format('"ell" indices', self.ell_indices)]
            s += ['{:18} : {}'.format('"emm" indices', self.emm_indices)]
            s += ['{:18} : {}'.format('Projection matrix', self.projection_matrix)]
            s += ['{:18} : {}'.format('Hash', hex(hash(self))[2:8])]
        s += ['-' * wdt]
        return '\n'.join(s)

    def _repr_html_(self) -> str:
        s = []
        s += ['<h3>SphericalHarmonics</h3>']
        s += ['<table border="1" class="dataframe">']
        s += ['<thead><tr><th style="text-align: left;">Field</th><th>Size</th><th>Data</th></tr></thead>']
        s += ['<tbody>']
        with np.printoptions(threshold=4, edgeitems=2, precision=2, linewidth=40):
            s += ['<tr><td style="text-align: left;">Maximum "ell"</td>']
            s += [f'<td>{1}</td><td>{self.ell_max}</td></tr>']
            s += ['<tr><td style="text-align: left;">"ell" indices</td>']
            s += [f'<td>{len(self.ell_indices)}</td><td>{self.ell_indices}</td></tr>']
            s += ['<tr><td style="text-align: left;">"emm" indices</td>']
            s += [f'<td>{len(self.emm_indices)}</td><td>{self.emm_indices}</td></tr>']
            s += ['<tr><td style="text-align: left;">Coefficient projection matrix</td>']
            s += [f'<td>{self.projection_matrix.shape}</td><td>{self.projection_matrix}</td></tr>']
            s += ['<tr><td style="text-align: left;">Hash</td>']
            s += [f'<td>{len(hex(hash(self)))}</td><td>{hex(hash(self))[2:8]}</td></tr>']
        s += ['</tbody>']
        s += ['</table>']
        return '\n'.join(s)



import h5py as h5
import logging
import numpy as np
from numpy.typing import NDArray
from scipy.io import loadmat
from scipy.spatial.transform import Rotation
from .geometry import Geometry
from .stack import Stack, Frame

logger = logging.getLogger(__name__)


class DataContainer:

    """
    Instances of this class represent data read from an input file in a format suitable for further analysis.
    The two core components are the :attr:`geometry` and the :attr:`stack`.
    The latter comprises a list of :class:`Frame <mumott.data_handling.stack.Frame>` instances,
    each of which corresponds to a single measurement.

    By default all data is read, which can be rather time consuming and unnecessary in some cases,
    e.g., when aligning data.
    In those cases, one can skip loading the actual measurements by setting :attr:`skip_data` to ``True``.
    The geometry information and supplementary information such as the diode data will still be read.

    Example
    -------
    The following code snippet illustrates the basic use of the :class:`DataContainer` class.

    First we create a :class:`DataContainer` instance, providing the path to the data file to be read.

    >>> from mumott.data_handling import DataContainer
    >>> dc = DataContainer('tests/test_full_circle.h5')

    One can then print a short summary of the content of the :class:`DataContainer` instance.

    >>> print(dc)
    ==========================================================================
                                  DataContainer
    --------------------------------------------------------------------------
    Corrected for transmission : False
    ...

    To access individual measurements we can use the :attr:`stack` attribute.
    The latter behaves like a list, where the elements of the list are
    :class:`Frame <mumott.data_handling.stack.Frame>` objects,
    each of which represents an individual measurement.
    We can print a summary of the content of the first frame.

    >>> print(dc.stack[0])
    --------------------------------------------------------------------------
                                      Frame
    --------------------------------------------------------------------------
    hash_data          : 475981
    hash_diode         : 40a75d
    hash_weights       : a213c4
    rotation           : [1. 0. 0.], [0. 1. 0.], [0. 0. 1.]
    j_offset           : -0.01
    k_offset           : 5.0
    --------------------------------------------------------------------------

    Parameters
    ----------
    data_path : str
        Path of the data file relative to the directory of execution.
    data_type : str, optional
        The type (or format) of the data file. Supported values are
        ``h5`` (default) for hdf5 format and ``mat`` for cSAXS Matlab format.
    skip_data : bool, optional
        If ``True``, will skip data from individual measurements when loading the file.
        This will result in a functioning :attr:`geometry` instance as well as
        :attr:`diode` and :attr:`weights` entries in the data stack, but
        :attr:`data` will be empty.
    """
    def __init__(self,
                 data_path: str,
                 data_type: str = 'h5',
                 skip_data: bool = False):
        self._correct_for_transmission_called = False
        self._stack = Stack()
        self._geometry_dictionary = dict()
        self._skip_data = skip_data
        if data_type == 'mat':
            self._matlab_to_stack(data_path)
        elif data_type == 'h5':
            self._h5_to_stack(data_path)
        else:
            raise ValueError(f'Unknown data_type: {data_type} for'
                             ' load_only_geometry=False.')

    def _h5_to_stack(self, file_path: str):
        """
        Internal method for loading data from hdf5 file.
        """
        h5_data = h5.File(file_path, 'r')
        projections = h5_data['projections']
        number_of_projections = len(projections)
        max_shape = (0, 0)
        for i in range(number_of_projections):
            p = projections[f'{i}']
            if 'diode' in p:
                max_shape = np.max((max_shape, p['diode'].shape), axis=0)
        for i in range(number_of_projections):
            p = projections[f'{i}']
            if 'diode' in p:
                diode = np.ascontiguousarray(np.copy(p['diode']).astype(np.float64))
                pad_sequence = np.array(((0, max_shape[0] - diode.shape[0]),
                                         (0, max_shape[1] - diode.shape[1]),
                                         (0, 0)))
                diode = np.pad(diode, pad_sequence[:-1])
            elif 'data' in p:
                diode = None
                pad_sequence = np.array(((0, max_shape[0] - p['data'].shape[0]),
                                         (0, max_shape[1] - p['data'].shape[1]),
                                         (0, 0)))
            else:
                pad_sequence = np.zeros((3, 2))
            if not self._skip_data:
                data = np.ascontiguousarray(np.copy(p['data']).astype(np.float64))
                data = np.pad(data, pad_sequence)
                if 'weights' in p:
                    weights = np.ascontiguousarray(
                        np.copy(p['weights']).astype(np.float64))[:, :, np.newaxis] * \
                        np.ones((1, 1, data.shape[-1])).astype(np.float64)
                else:
                    weights = np.ones_like(data)
            else:
                data = None
                if 'weights' in p:
                    weights = np.ascontiguousarray(np.copy(p['weights']).astype(np.float64))
                    weights = np.pad(weights, pad_sequence[:weights.ndim])
                else:
                    if diode is None:
                        weights = None
                    else:
                        weights = np.ones_like(diode)
                        weights = np.pad(weights, pad_sequence[:weights.ndim])
            # Look for rotation information and load if available
            if 'rot_mat' in p:
                rotation = p['rot_mat'][...]
                if i == 0:
                    logger.info('Rotation matrices were loaded from the input file.')
            elif 'rotations' in p and 'tilts' in p:
                if 'inner_axis' in p and 'outer_axis' in p:
                    R_rot = Rotation.from_vector(p['rotations'] * p['inner_axis']).as_matrix()
                    R_tilt = Rotation.from_vector(p['tilts'] * p['outer_axis']).as_matrix()
                    if i == 0:
                        logger.info('Rotation matrices were generated from rotation and tilt angles,'
                                    ' along with inner and outer rotation axis vectors.'
                                    ' Rotation and tilt angles assumed to be in radians.')
                else:
                    R_rot = self._Rz(-p['rotations'][0])
                    R_tilt = self._Rx(p['tilts'][0])
                    if i == 0:
                        logger.info('Rotation matrices were generated from rotation and tilt angles.'
                                    ' Rotations were assumed to be about the z-axis and tilts about x-axis'
                                    ' with angles assumed to be in radians.')
                rotation = R_tilt @ R_rot
            else:
                rotation = np.eye(3)
                if i == 0:
                    logger.info('No rotation information found.')
            j_offset = p['offset_j'][0]
            k_offset = p['offset_k'][0]
            j_offset -= pad_sequence[0, 1] * 0.5
            k_offset -= pad_sequence[1, 1] * 0.5
            frame = Frame(data=data,
                          diode=diode,
                          weights=weights,
                          rotation=rotation,
                          j_offset=j_offset,
                          k_offset=k_offset)
            self._stack.append(frame)
        if not self._skip_data:
            self._stack.geometry.detector_angles = np.copy(h5_data['detector_angles'])
        if 'volume_shape' in h5_data.keys():
            self._stack.geometry.volume_shape = np.copy(h5_data['volume_shape'])
        else:
            self._stack.geometry.volume_shape = np.array(max_shape)[[0, 0, 1]]
        # Load sample geometry information
        if 'p_direction_0' in h5_data.keys():  # TODO check for orthogonality, normality
            self._stack.geometry.p_direction_0 = np.copy(h5_data['p_direction_0'][...])
            self._stack.geometry.j_direction_0 = np.copy(h5_data['j_direction_0'][...])
            self._stack.geometry.k_direction_0 = np.copy(h5_data['k_direction_0'][...])
            logger.info('Sample geometry loaded from file.')
        else:
            logger.info('No sample geometry information was found. Default mumott geometry assumed.')

        # Load detector geometry information
        if 'detector_direction_origin' in h5_data.keys():  # TODO check for orthogonality, normality
            self._stack.geometry.detector_direction_origin = np.copy(
                h5_data['detector_direction_origin'][...])
            self._stack.geometry.detector_direction_positive_90 = np.copy(
                h5_data['detector_direction_positive_90'][...])
            logger.info('Detector geometry loaded from file.')
        else:
            logger.info('No detector geometry information was found. Default mumott geometry assumed.')

    def _matlab_to_stack(self, file_path: str):
        """
        Internal method for loading data from Matlab file.
        """
        try:
            matlab_data = loadmat(file_path)
            is_v73 = False
        except Exception as e:
            if 'v7.3' in str(e):
                logger.info('Matlab file version is v7.3, using h5py to load.')
                matlab_data = h5.File(file_path)
                is_v73 = True
            else:
                logger.warning('scipy.io.loadmat failed for unidentified reasons.'
                               ' Trying h5py.')
                matlab_data = h5.File(file_path)
                is_v73 = True
                logger.warning(
                    'The following exception was raised during execution of scipy.io.loadmat:\n' + str(e) +
                    '\nPlease proceed with caution.')
        number_of_projections = matlab_data['projection']['data'].size
        any_tomo_axis_x = False

        def load_one_frame(i: int) -> tuple:
            if is_v73:
                if not self._skip_data:
                    data = np.array(matlab_data[matlab_data['projection']['data'][i, 0]],
                                    copy=True).squeeze().astype(np.float64).T
                else:
                    data = None
                diode = np.array(matlab_data[matlab_data['projection']['diode'][i, 0]],
                                 copy=True).squeeze().astype(np.float64).T
                temp_rot_x = np.array(matlab_data[matlab_data['projection']['rot_x'][i, 0]],
                                      copy=True).squeeze().astype(np.float64)
                temp_rot_y = np.array(matlab_data[matlab_data['projection']['rot_y'][i, 0]],
                                      copy=True).squeeze().astype(np.float64)
                temp_dx = np.array(matlab_data[matlab_data['projection']['dx'][i, 0]],
                                   copy=True).squeeze().astype(np.float64)
                temp_dy = np.array(matlab_data[matlab_data['projection']['dy'][i, 0]],
                                   copy=True).squeeze().astype(np.float64)
                window_mask = np.array(matlab_data[matlab_data['projection']['window_mask'][i, 0]],
                                       copy=True).squeeze().astype(np.float64).T
                tomo_axis_x = round(matlab_data[matlab_data['projection']['par'][0, 0]]['tomo_axis_x'][0, 0])
            else:
                if not self._skip_data:
                    data = np.copy(matlab_data['projection']['data'][0, i]).astype(np.float64)
                else:
                    data = None
                diode = np.copy(matlab_data['projection']['diode'][0, i]).astype(np.float64)
                temp_rot_x = np.copy(matlab_data['projection']['rot_x'][0, i].squeeze()).astype(np.float64)
                temp_rot_y = np.copy(matlab_data['projection']['rot_y'][0, i].squeeze()).astype(np.float64)
                temp_dx = np.copy(matlab_data['projection']['dx'][0, i].squeeze()).astype(np.float64)
                temp_dy = np.copy(matlab_data['projection']['dy'][0, i].squeeze()).astype(np.float64)
                window_mask = np.copy(matlab_data['projection']['window_mask'][0, i]).astype(np.float64)
                tomo_axis_x = round(matlab_data['projection']['par'][0, 0]['tomo_axis_x'][0, 0][0, 0])
            return (np.ascontiguousarray(data), np.ascontiguousarray(diode),
                    temp_rot_x, temp_rot_y, temp_dx, temp_dy,
                    np.ascontiguousarray(window_mask), tomo_axis_x)
        max_shape = (0, 0)
        for i in range(number_of_projections):
            diode = load_one_frame(i)[1]
            max_shape = np.max((max_shape, diode.shape[:2]), axis=0)

        for i in range(number_of_projections):
            data, diode, temp_rot_x, temp_rot_y, temp_dx, temp_dy, window_mask, tomo_axis_x = \
                load_one_frame(i)
            if not self._skip_data:
                weights = window_mask[..., np.newaxis] * \
                    np.ones((1, 1, data.shape[-1]))
            else:
                weights = window_mask
            # Pad to make all frames the same shape.
            pad_sequence = np.array(((0, max_shape[0] - diode.shape[0]),
                                    (0, max_shape[1] - diode.shape[1]),
                                    (0, 0)))
            if not self._skip_data:
                data = np.pad(data, pad_sequence)
            diode = np.pad(diode, pad_sequence[:-1])
            weights = np.pad(weights, pad_sequence[:weights.ndim])
            if tomo_axis_x:
                any_tomo_axis_x = True
                rotation = \
                    self._Ry(-temp_rot_y * np.pi / 180) @ self._Rx(-temp_rot_x * np.pi / 180)
            else:
                rotation = \
                    self._Rx(temp_rot_x * np.pi / 180) @ self._Ry(temp_rot_y * np.pi / 180)
            # Append frame to stack
            j_offset = temp_dy - pad_sequence[0, 1] * 0.5
            k_offset = temp_dx - pad_sequence[1, 1] * 0.5
            frame = Frame(data=data,
                          diode=diode,
                          weights=weights,
                          rotation=rotation,
                          j_offset=j_offset,
                          k_offset=k_offset)
            self._stack.append(frame)
        number_of_segments = self._stack.data.shape[-1]
        if not self._skip_data:
            if is_v73:
                self._stack.geometry.detector_angles = np.array(matlab_data[matlab_data['projection']
                                                                ['integ'][0, 0]]
                                                                ['phi_det'][:number_of_segments, 0],
                                                                copy=True) * np.pi / 180
                matlab_data.close()
            else:
                self._stack.geometry.detector_angles = np.copy(
                    matlab_data['projection']['integ'][0, 0]
                    ['phi_det'][0, 0][0, :number_of_segments]) * np.pi / 180
        self._stack.geometry.p_direction_0 = np.array([0., 0., 1.0])
        self._stack.geometry.j_direction_0 = np.array([0., 1.0, 0.])
        self._stack.geometry.k_direction_0 = np.array([1.0, 0., 0.])
        self._stack.geometry.detector_direction_origin = np.array([1.0, 0, 0.0])
        self._stack.geometry.detector_direction_positive_90 = np.array([0, 1.0, 0])
        if any_tomo_axis_x:
            self._stack.geometry.volume_shape = np.array(max_shape)[[1, 0, 0]]
        else:
            self._stack.geometry.volume_shape = np.array(max_shape)[[1, 0, 1]]

    def __len__(self) -> int:
        """
        Length of the :attr:`stack <mumott.data_handling.stack.Stack>`
        attached to this :class:`DataContainer` instance.
        """
        return len(self._stack)

    def append(self, f: Frame) -> None:
        """
        Appends a :class:`Frame <mumott.data_handling.stack.Frame>` to the :attr:`stack`
        attached to this :class:`DataContainer` instance.
        """
        self._stack.append(f)

    @property
    def stack(self) -> Stack:
        """ The stack, containing data and geometry. """
        return self._stack

    @stack.setter
    def stack(self, val) -> None:
        self._stack = val

    @property
    def geometry(self) -> Geometry:
        """ Container of geometry information. """
        return self._stack.geometry

    @property
    def data(self) -> NDArray[np.float64]:
        """
        The data in the :attr:`stack` object
        attached to this :class:`DataContainer` instance.
        """
        return self._stack.data

    @property
    def diode(self) -> NDArray[np.float64]:
        """
        The diode data in the :attr:`stack` object
        attached to this :class:`DataContainer` instance.
        """
        return self._stack.diode

    @property
    def weights(self) -> NDArray[np.float64]:
        """
        The weights in the :attr:`stack` object
        attached to this :class:`DataContainer` instance.
        """
        return self._stack.weights

    def correct_for_transmission(self) -> None:
        """
        Applies correction from the input provided in the :attr:`diode
        <mumott.data_handling.stack.Frame>` field.  Should
        only be used if this correction has not already been applied.
        """
        if self._correct_for_transmission_called:
            logger.info(
                'DataContainer.correct_for_transmission() has been called already.'
                ' The correction has been applied previously, and the repeat call is ignored.')
            return

        data = self._stack.data / self._stack.diode[..., np.newaxis]

        for i, f in enumerate(self._stack):
            f.data = data[i]
        self._correct_for_transmission_called = True

    def _Rx(self, angle: float) -> NDArray[float]:
        """ Generate a rotation matrix for rotations around
        the x-axis, following the convention that vectors
        have components ordered ``(x, y, z)``.

        Parameters
        ----------
        angle
            The angle of the rotation.

        Returns
        -------
        R
            The rotation matrix.

        Notes
        -----
        For a vector ``v`` with shape ``(..., 3)`` and a rotation angle :attr:`angle`,
        ``np.einsum('ji, ...i', _Rx(angle), v)`` rotates the vector around the
        ``x``-axis by :attr:`angle`. If the
        coordinate system is being rotated, then
        ``np.einsum('ij, ...i', _Rx(angle), v)`` gives the vector in the
        new coordinate system.
        """
        return Rotation.from_euler('X', angle).as_matrix()

    def _Ry(self, angle: float) -> NDArray[float]:
        """ Generate a rotation matrix for rotations around
        the y-axis, following the convention that vectors
        have components ordered ``(x, y, z)``.

        Parameters
        ----------
        angle
            The angle of the rotation.

        Returns
        -------
        R
            The rotation matrix.

        Notes
        -----
        For a vector ``v`` with shape ``(..., 3)`` and a rotation angle ``angle``,
        ``np.einsum('ji, ...i', _Ry(angle), v)`` rotates the vector around the
        For a vector ``v`` with shape ``(..., 3)`` and a rotation angle :attr:`angle`,
        ``np.einsum('ji, ...i', _Ry(angle), v)`` rotates the vector around the
        ``y``-axis by :attr:`angle`. If the
        coordinate system is being rotated, then
        ``np.einsum('ij, ...i', _Ry(angle), v)`` gives the vector in the
        new coordinate system.
        """
        return Rotation.from_euler('Y', angle).as_matrix()

    def _Rz(self, angle: float) -> NDArray[float]:
        """ Generate a rotation matrix for rotations around
        the z-axis, following the convention that vectors
        have components ordered ``(x, y, z)``.

        Parameters
        ----------
        angle
            The angle of the rotation.

        Returns
        -------
        R
            The rotation matrix.

        Notes
        -----
        For a vector ``v`` with shape ``(..., 3)`` and a rotation angle :attr:`angle`,
        ``np.einsum('ji, ...i', _Rz(angle), v)`` rotates the vector around the
        ``z``-axis by :attr:`angle`. If the
        coordinate system is being rotated, then
        ``np.einsum('ij, ...i', _Rz(angle), v)`` gives the vector in the
        new coordinate system.
        """
        return Rotation.from_euler('Z', angle).as_matrix()

    def _get_str_representation(self, max_lines=50) -> str:
        """ Retrieves a string representation of the object with specified
        maximum number of lines.

        Parameters
        ----------
        max_lines
            The maximum number of lines to return.
        """
        wdt = 74
        s = []
        s += ['=' * wdt]
        s += ['DataContainer'.center(wdt)]
        s += ['-' * wdt]
        s += ['{:26} : {}'.format('Corrected for transmission', self._correct_for_transmission_called)]
        truncated_s = []
        leave_loop = False
        while not leave_loop:
            line = s.pop(0).split('\n')
            for split_line in line:
                if split_line != '':
                    truncated_s += [split_line]
                if len(truncated_s) > max_lines - 2:
                    if split_line != '...':
                        truncated_s += ['...']
                    if split_line != ('=' * wdt):
                        truncated_s += ['=' * wdt]
                    leave_loop = True
                    break
            if len(s) == 0:
                leave_loop = True
        truncated_s += ['=' * wdt]
        return '\n'.join(truncated_s)

    def __str__(self) -> str:
        return self._get_str_representation()

    def _get_html_representation(self, max_lines=25) -> str:
        """ Retrieves an html representation of the object with specified
        maximum number of lines.

        Parameters
        ----------
        max_lines
            The maximum number of lines to return.
        """
        s = []
        s += ['<h3>DataContainer</h3>']
        s += ['<table border="1" class="dataframe">']
        s += ['<thead><tr><th style="text-align: left;">Field</th><th>Size</th></tr></thead>']
        s += ['<tbody>']
        s += ['<tr><td style="text-align: left;">Number of projections</td>']
        s += [f'<td>{len(self._stack)}</td></tr>']
        s += ['<tr><td style="text-align: left;">Corrected for transmission</td>']
        s += [f'<td>{self._correct_for_transmission_called}</td></tr>']
        s += ['</tbody>']
        s += ['</table>']
        truncated_s = []
        line_count = 0
        leave_loop = False
        while not leave_loop:
            line = s.pop(0).split('\n')
            for split_line in line:
                truncated_s += [split_line]
                if '</tr>' in split_line:
                    line_count += 1
                    # Catch if last line had ellipses
                    last_tr = split_line
                if line_count > max_lines - 1:
                    if last_tr != '<tr><td style="text-align: left;">...</td></tr>':
                        truncated_s += ['<tr><td style="text-align: left;">...</td></tr>']
                    truncated_s += ['</tbody>']
                    truncated_s += ['</table>']
                    leave_loop = True
                    break
            if len(s) == 0:
                leave_loop = True
        return '\n'.join(truncated_s)

    def _repr_html_(self) -> str:
        return self._get_html_representation()



import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterator, Tuple

import numpy as np

from ..loss_functions.base_loss_function import LossFunction

logger = logging.getLogger(__name__)


class Optimizer(ABC):

    """This is the base class from which specific optimizers are derived.
    """

    def __init__(self,
                 loss_function: LossFunction,
                 **kwargs: Dict[str, Any]):
        self._loss_function = loss_function
        # empty kwargs automatically yields empty dict
        self._options = kwargs

    # set, get and iter methods for kwargs interface
    def __setitem__(self, key: str, val: Any) -> None:
        """ Sets options akin to ``**kwargs`` during initialization. Allows
        access to instance as a dictionary.

        Parameters
        ----------
        key
            The key used in ``dict``-like interface to instance.
        val
            The value to store in association with ``key``.
        """
        if key not in self._options.keys():
            logger.info(f'Key {key} added to options with value {val}.')
        self._options[key] = val

    def __getitem__(self, key: str) -> Any:
        """ Sets options akin to ``**kwargs`` during initialization. Allows
        access to instance as a dictionary.

        Parameters
        ----------
        key
            The key used in ``dict``-like interface to instance.
        val
            The value to store in association with ``key``.
        Returns
            Value stored in association with ``key``
        """
        if key in self._options.keys():
            return self._options[key]
        else:
            raise KeyError(f'Unrecognized key: {key}')

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        """ Allows casting as dict and use as an iterator. """
        for t in self._options.items():
            yield t

    @abstractmethod
    def optimize(self) -> Dict:
        """ Function for executing the optimization. Should return a ``dict`` of the
        optimization results. """
        pass

    def __str__(self) -> str:
        wdt = 74
        s = []
        s += ['=' * wdt]
        s += [__class__.__name__.center(wdt)]
        s += ['-' * wdt]
        with np.printoptions(threshold=4, precision=5, linewidth=60, edgeitems=1):
            s += ['{:18} : {}'.format('LossFunction', self._loss_function.__class__.__name__)]
            s += ['{:18} : {}'.format('hash', hex(hash(self))[2:8])]
            for key, value in self._options.items():
                s += ['{:18} : {}'.format(f'option[{key}]', value)]
        s += ['-' * wdt]
        return '\n'.join(s)

    def _repr_html_(self) -> str:
        s = []
        s += [f'<h3>{__class__.__name__}</h3>']
        s += ['<table border="1" class="dataframe">']
        s += ['<thead><tr><th style="text-align: left;">Field</th><th>Size</th><th>Data</th></tr></thead>']
        s += ['<tbody>']
        with np.printoptions(threshold=4, edgeitems=2, precision=2, linewidth=40):
            s += ['<tr><td style="text-align: left;">LossFunction</td>']
            s += [f'<td>{1}</td><td>{self._loss_function.__class__.__name__}</td></tr>']
            h = hex(hash(self))
            s += ['<tr><td style="text-align: left;">Hash</td>']
            s += [f'<td>{len(h)}</td><td>{h[2:8]}</td></tr>']
            for key, value in self._options.items():
                s += [f'<tr><td style="text-align: left;">options[{key}]</td>']
                s += [f'<td>{1}</td><td>{value}</td></tr>']
        s += ['</tbody>']
        s += ['</table>']
        return '\n'.join(s)



import pytest
import numpy as np

from mumott import Geometry
from mumott.data_handling.geometry import GeometryTuple


tuple_list = [GeometryTuple(rotation=np.eye(3), j_offset=0.3, k_offset=0.7),
              GeometryTuple(rotation=np.eye(3) * 0.732, j_offset=0.3, k_offset=0.7),
              GeometryTuple(rotation=-np.eye(3), j_offset=0.3, k_offset=0.7)]

rotation_array = np.array([[[0.955336, 0.190379, 0.226026],
                           [0., 0.764842, -0.644218],
                           [-0.2955, 0.615445,  0.730682]],
                          [[0.955336, 0.190379,  0.226026],
                           [0., 0.764842, -0.644218],
                           [-0.29552, 0.615445,  0.730682]]])

rotation_list = [np.array([[0.955336, 0.190379, 0.226026],
                          [0., 0.764842, -0.644218],
                          [-0.2955, 0.615445,  0.730682]]),
                 np.array([[0.955336, 0.190379,  0.226026],
                          [0., 0.764842, -0.644218],
                          [-0.29552, 0.615445,  0.730682]])]

j_list = [0.7, 0.8, 0.4]
k_list = [0.3, 0.5, 0.1]

j_array = np.array([2.1, -1, 0.2322])
k_array = np.array([2.1, -1, 0.2322])


@pytest.mark.parametrize('test_list', [(tuple_list)])
def test_input_output(test_list, tmp_path):
    d = tmp_path / 'sub'
    d.mkdir()
    p = d / 'test.geo'
    g = Geometry()
    for t in tuple_list:
        g.append(t)
    for a, b in zip(g, tuple_list):
        assert np.allclose(a.rotation, b.rotation)
        assert np.isclose(a.j_offset, b.j_offset)
        assert np.isclose(a.k_offset, b.k_offset)
    g.write(p)
    k = Geometry()
    k.read(p)
    for a, b in zip(k, tuple_list):
        assert np.allclose(a.rotation, b.rotation)
        assert np.isclose(a.j_offset, b.j_offset)
        assert np.isclose(a.k_offset, b.k_offset)
    assert hash(k) == hash(g)
    m = Geometry(p)
    for a, b in zip(m, tuple_list):
        assert np.allclose(a.rotation, b.rotation)
        assert np.isclose(a.j_offset, b.j_offset)
        assert np.isclose(a.k_offset, b.k_offset)
    assert hash(g) == hash(m)


@pytest.mark.parametrize('test_list,test_array', [(rotation_list, rotation_array)])
def test_rotation_list_array(test_list, test_array):
    g = Geometry()
    print(test_array)
    print(test_list)
    g.rotations = test_array
    assert np.allclose(g.rotations_as_array, test_array)
    for a, b in zip(g, test_array):
        assert np.allclose(a.rotation, b)
    for a, b in zip(g.rotations, test_array):
        assert np.allclose(a, b)

    g.rotations = test_list
    assert np.allclose(g.rotations_as_array, test_list)
    for a, b in zip(g, test_list):
        assert np.allclose(a.rotation, b)
    for a, b in zip(g.rotations, test_list):
        assert np.allclose(a, b)


@pytest.mark.parametrize('test_list,test_array', [(j_list, j_array)])
def test_j_list_array(test_list, test_array):
    g = Geometry()
    g.j_offsets = test_array
    assert np.allclose(g.j_offsets_as_array, test_array)
    for a, b in zip(g, test_array):
        assert np.allclose(a.j_offset, b)
    for a, b in zip(g.j_offsets, test_array):
        assert np.allclose(a, b)

    g.j_offsets = test_list
    assert np.allclose(g.j_offsets, test_list)
    for a, b in zip(g, test_list):
        assert np.allclose(a.j_offset, b)
    for a, b in zip(g.j_offsets, test_list):
        assert np.allclose(a, b)


@pytest.mark.parametrize('test_list,test_array',  [(k_list, k_array)])
def test_k_list_array(test_list, test_array):
    g = Geometry()
    g.k_offsets = test_array
    assert np.allclose(g.k_offsets_as_array, test_array)
    for a, b in zip(g, test_array):
        assert np.allclose(a.k_offset, b)
    for a, b in zip(g.k_offsets, test_array):
        assert np.allclose(a, b)

    g.k_offsets = test_list
    assert np.allclose(g.k_offsets, test_list)
    for a, b in zip(g, test_list):
        assert np.allclose(a.k_offset, b)
    for a, b in zip(g.k_offsets, test_list):
        assert np.allclose(a, b)


def test_empty_hash():
    g = Geometry()
    assert str(hash(g))[:6] == '135563'
    assert g.hash_rotations[:6] == '786a02'
    assert g.hash_j_offsets[:6] == '786a02'
    assert g.hash_k_offsets[:6] == '786a02'


hash_list = ['119687', '134744', '334171']


@pytest.mark.parametrize('tup,hsh', [t for t in zip(tuple_list, hash_list)])
def test_gm_tuple_hash(tup, hsh):
    assert str(hash(tup))[:6] == hsh
